from __future__ import annotations

import hashlib
import logging
from collections.abc import Iterable
from datetime import time
from typing import TYPE_CHECKING, TypeVar

from ..cos.objects.base import PdfHexString, PdfName, PdfNull, PdfObject, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.objects.stream import PdfStream

if TYPE_CHECKING:
    from pdfnaut.cos.parser import PdfParser


LOGGER = logging.getLogger(__name__)
Placeholder = type("Placeholder", (), {})


def get_value_from_bytes(contents: PdfHexString | bytes) -> bytes:
    """Returns the decoded value of ``contents`` if it is an instance of
    :class:`.PdfHexString`, otherwise returns ``contents``."""
    return contents.value if isinstance(contents, PdfHexString) else contents


R = TypeVar("R")


def ensure_object(obj: PdfReference[R] | R) -> R:
    """Resolves ``obj`` to a direct object if ``obj`` is an instance of
    :class:`.PdfReference`. Otherwise, returns ``obj`` as is."""
    if isinstance(obj, PdfReference):
        return obj.get()

    return obj


def get_closest(values: Iterable[int], target: int) -> int:
    """Returns the integer in ``values`` closest to ``target``."""
    return min(values, key=lambda offset: abs(offset - target))


def generate_file_id(filename: str, content_size: int) -> PdfHexString:
    """Generates a file identifier using ``filename`` and ``content_size`` as
    described in § 14.4, "File identifiers".

    File identifiers are values that uniquely separate a revision of a document
    from another. The file identifier is generated using the same information
    specified in the standard, that is, the current time, the file path and
    the file size in bytes.
    """

    id_digest = hashlib.md5(time().isoformat("auto").encode())
    id_digest.update(filename.encode())
    id_digest.update(str(content_size).encode())

    return PdfHexString(id_digest.hexdigest().encode())


def is_page_or_page_tree(obj: PdfObject | PdfStream) -> bool:
    """Reports whether an object ``obj`` is a page object or a page tree node."""

    if not isinstance(obj, PdfDictionary) or "Type" not in obj:
        return False

    if not isinstance(tp := obj["Type"], PdfName) or tp.value not in [b"Page", b"Pages"]:
        return False

    return True


def copy_object(obj: PdfObject | PdfStream) -> PdfObject | PdfStream:
    """Performs a deep copy of a PDF object ``obj``. Returns the copied object.

    Deep copying works by creating a new object for the container then adding a copy
    of each element it contains into the new object.

    Numbers, literal strings, booleans, and the null object are not copied and are
    returned as is. Unlike :meth:`.clone_in_document`, when a reference is found,
    it is simply copied into the object without modifying the referred object.
    """

    if isinstance(obj, PdfDictionary):
        kv = PdfDictionary()
        for key, value in obj.data.items():
            kv.data[key] = copy_object(value)
        return kv
    elif isinstance(obj, PdfStream):
        return PdfStream(copy_object(obj.details), obj.raw, copy_object(obj._crypt_params))
    elif isinstance(obj, PdfArray):
        arr = PdfArray()
        for value in obj.data:
            arr.data.append(copy_object(value))
        return arr
    elif isinstance(obj, PdfName):
        return PdfName(obj.value)
    elif isinstance(obj, PdfHexString):
        return PdfHexString(obj.raw)
    elif isinstance(obj, PdfReference):
        return PdfReference(obj.object_number, obj.generation)

    return obj


def clone_into_document(
    dest: PdfParser, root: PdfObject | PdfStream, *, ignore_keys: list[str] | None = None
) -> PdfObject | PdfStream:
    """Clones an object ``root`` and its contents into document ``dest``. Returns
    the cloned object.

    If the root object is a dictionary and the ``ignore_keys`` argument is provided,
    those keys will be ignored when cloning the root object.

    Cloning of an object is performed by deep-copying each element contained in it.
    When a reference is found, it is determined whether it is suitable for cloning
    into the document.

    A reference is determined suitable for cloning if it does not refer back to the
    ``root`` object. If it is unsuitable, a placeholder is added if the reference is
    ``root`` itself. If the reference may point back to the object (such as the
    reference being for a page tree), it is nulled.

    If the reference is suitable, its contents are added into the document and the new
    references replaces the old reference in the object.
    """

    if ignore_keys is None:
        ignore_keys = []

    cloned_map = {}
    references = set()

    def inner(obj: PdfObject | PdfStream) -> type[Placeholder] | PdfObject | PdfStream:
        if obj in cloned_map:
            # object is already cloned
            return cloned_map[obj]

        if isinstance(obj, PdfReference):
            referred = obj.get()

            if referred is root:
                # object refers to our origin object. in which case, simply set
                # a placeholder for later processing
                return Placeholder

            if is_page_or_page_tree(referred):
                # avoid going to pages or anything that might lead us to the page tree
                LOGGER.warning("object %s cannot be reliably copied and has been set to null.", obj)
                return PdfNull()

            cloned_direct = inner(referred)
            cloned_map[obj] = dest.objects.add(cloned_direct)
            references.add(cloned_map[obj])
            return cloned_map[obj]
        elif isinstance(obj, PdfDictionary):
            kv = PdfDictionary()
            cloned_map[obj] = kv
            for key, value in obj.data.items():
                if obj is root and key in ignore_keys:
                    continue

                kv.data[key] = inner(value)
            return kv
        elif isinstance(obj, PdfStream):
            stm = PdfStream(inner(obj.details), obj.raw, inner(obj._crypt_params))
            cloned_map[obj] = stm
            return stm
        elif isinstance(obj, PdfArray):
            arr = PdfArray()
            cloned_map[obj] = arr
            for value in obj.data:
                arr.data.append(inner(value))
            return arr
        elif isinstance(obj, PdfName):
            return PdfName(obj.value)
        elif isinstance(obj, PdfHexString):
            return PdfHexString(obj.raw)

        return obj

    cloned = inner(root)

    def replace_placeholders(obj: PdfObject | type[Placeholder]) -> PdfObject:
        if obj is Placeholder:
            return dest.objects.add(cloned)
        elif isinstance(obj, PdfArray):
            return PdfArray(replace_placeholders(it) for it in obj.data)
        elif isinstance(obj, PdfDictionary):
            return PdfDictionary({key: replace_placeholders(val) for key, val in obj.data.items()})
        elif isinstance(obj, PdfStream):
            obj.details = replace_placeholders(obj.details)
            obj._crypt_params = replace_placeholders(obj._crypt_params)
            return obj

        return obj

    final = replace_placeholders(cloned)
    for ref in references:
        direct = dest.objects[ref.object_number]
        dest.objects[ref.object_number] = replace_placeholders(direct)

    return final



from __future__ import annotations

from typing import Literal, Union

from pdfnaut.common.dictmodels import dictmodel, field
from pdfnaut.cos.objects import PdfDictionary

NumberingStyle = Literal["D", "R", "r", "A", "a"]


def test_basic_dictmodel() -> None:
    @dictmodel()
    class Point(PdfDictionary):
        x: int
        y: int
        z: Union[int, None] = None

    assert hasattr(Point, "__accessors__")

    p = Point(x=10, y=20)
    assert p.data == {"X": 10, "Y": 20}

    p.z = -10
    assert "Z" in p.data and p["Z"] == -10


def test_dictmodel_with_defaults() -> None:
    @dictmodel()
    class PageLabelScheme(PdfDictionary):
        style: Union[NumberingStyle, None] = field("S", default=None)
        prefix: Union[str, None] = field("P", default=None)
        start: int = field("St", default=1)

    scheme = PageLabelScheme(style="D")
    assert scheme.style == "D" and scheme.prefix is None and scheme.start == 1

    scheme["St"] = 10
    assert scheme.start == 10


def test_inherited_dictmodel() -> None:
    @dictmodel()
    class Point2D(PdfDictionary):
        x: int
        y: int

    @dictmodel()
    class Point3D(Point2D):
        z: int

    p2 = Point2D(10, 20)
    assert p2.data == {"X": 10, "Y": 20}

    p3 = Point3D(10, 20, 30)
    assert p3.data == {"X": 10, "Y": 20, "Z": 30}



from __future__ import annotations

from collections import defaultdict
from typing import Any, Literal

from ..cos.objects.base import PdfComment, PdfHexString, PdfName, PdfNull, PdfObject, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.objects.stream import PdfStream
from ..cos.objects.xref import (
    CompressedXRefEntry,
    FreeXRefEntry,
    InUseXRefEntry,
    PdfXRefEntry,
    PdfXRefSection,
    PdfXRefSubsection,
)
from ..exceptions import PdfWriteError
from .tokenizer import STRING_ESCAPE


def serialize_comment(comment: PdfComment) -> bytes:
    return b"%" + comment.value


def serialize_null(_) -> bytes:
    return b"null"


def serialize_bool(boolean: bool) -> bytes:
    return b"true" if boolean else b"false"


def serialize_literal_string(byte_str: bytes, *, keep_ascii: bool = False) -> bytes:
    output = bytearray()
    escape = {v: k for k, v in STRING_ESCAPE.items()}

    for char in byte_str:
        char = char.to_bytes(1, "big")

        if (esc := escape.get(char)) is not None:
            # character needs to be escaped
            output += esc
        elif keep_ascii and not char.isascii():
            # octal \ddd notation
            output += rf"\{ord(char):0>3o}".encode()
        else:
            # character does not need escaping
            output += char

    return b"(" + output + b")"


def serialize_name(name: PdfName) -> bytes:
    output = b"/"

    for char in name.value:
        char = char.to_bytes(1, "big")
        if char.isalnum():
            output += char
        else:
            output += rf"#{ord(char):x}".encode()

    return output


def serialize_hex_string(string: PdfHexString) -> bytes:
    return b"<" + string.raw + b">"


def serialize_indirect_ref(reference: PdfReference) -> bytes:
    return f"{reference.object_number} {reference.generation} R".encode()


def serialize_numeric(number: int | float) -> bytes:
    return str(number).encode()


def serialize_array(array: PdfArray) -> bytes:
    return b"[" + b" ".join(serialize(item) for item in array.data) + b"]"


def serialize_dictionary(dictionary: PdfDictionary) -> bytes:
    items = []
    for key, val in dictionary.data.items():
        items.append(serialize(PdfName(key.encode())))
        items.append(serialize(val))

    return b"<<" + b" ".join(items) + b">>"


def serialize_stream(stream: PdfStream, *, eol: bytes) -> bytes:
    return b"".join(
        [
            serialize_dictionary(stream.details) + eol,
            b"stream" + eol,
            stream.raw + eol,
            b"endstream",
        ]
    )


def serialize(
    object_: PdfObject | PdfStream | PdfComment, *, params: dict[str, Any] | None = None
) -> bytes:
    if params is None:
        params = {}

    if isinstance(object_, PdfComment):
        return serialize_comment(object_)
    elif isinstance(object_, PdfName):
        return serialize_name(object_)
    elif isinstance(object_, bytes):
        return serialize_literal_string(object_, keep_ascii=params.get("keep_ascii", False))
    elif isinstance(object_, bool):
        return serialize_bool(object_)
    elif isinstance(object_, PdfNull):
        return serialize_null(object_)
    elif isinstance(object_, PdfHexString):
        return serialize_hex_string(object_)
    elif isinstance(object_, PdfReference):
        return serialize_indirect_ref(object_)
    elif isinstance(object_, (int, float)):
        return serialize_numeric(object_)
    elif isinstance(object_, PdfArray):
        return serialize_array(object_)
    elif isinstance(object_, PdfDictionary):
        return serialize_dictionary(object_)
    elif isinstance(object_, PdfStream):
        return serialize_stream(object_, eol=params["eol"])

    raise PdfWriteError(f"Cannot serialize type {type(object_)}")


class PdfSerializer:
    """A serializer that is able to produce a valid PDF document.

    Arguments:
        eol (bytes, optional):
            The end-of-line to be used when serializing (CR, LF, or CRLF). Defaults to CRLF.
    """

    def __init__(self, *, eol: Literal[b"\r\n", b"\r", b"\n"] = b"\r\n") -> None:
        self.content_lines: list[bytes] = []
        self.eol = eol

        self.objects: dict[tuple[int, int], PdfObject | PdfStream] = {}

    @property
    def content(self) -> bytes:
        """The serialized content to be written."""
        return b"".join(self.content_lines)

    def write_header(self, version: str, *, with_binary_marker: bool = True) -> None:
        """Appends the PDF file header to the document (see § 7.5.2, "File Header").

        Arguments:
            version (str):
                A string representing the version of the PDF file.

            with_binary_marker (bool, optional):
                Whether to also append the recommended binary marker. Defaults to True.
        """

        comment = PdfComment(f"PDF-{version}".encode())
        self.content_lines.append(serialize_comment(comment) + self.eol)

        if with_binary_marker:
            marker = PdfComment(b"\xee\xe1\xf5\xf4")
            self.content_lines.append(serialize_comment(marker) + self.eol)

    def write_object(
        self, reference: PdfReference | tuple[int, int], contents: PdfObject | PdfStream
    ) -> int:
        """Appends an indirect object to the document.

        Arguments:
            reference (PdfReference | tuple[int, int]):
                The object number and generation to which the object should be assigned.

            contents (PdfObject | PdfStream):
                The contents to associate with the reference.

        Returns:
            The offset where the indirect object starts.
        """
        if isinstance(reference, tuple):
            reference = PdfReference(*reference)

        offset = len(self.content)

        self.content_lines.extend(
            [
                f"{reference.object_number} {reference.generation} obj".encode() + self.eol,
                serialize(contents, params={"eol": self.eol}) + self.eol,
                b"endobj" + self.eol,
            ]
        )

        return offset

    def generate_xref_section(
        self, rows: list[tuple[int, PdfXRefEntry]]
    ) -> list[PdfXRefSubsection]:
        """Generates a cross-reference section from a list of ``rows``.

        Each row consists of a two-element tuple containing the object number of the
        XRef entry and the entry itself. The object numbers will determine the amount
        of subsections created and the entries within them.

        The output is a list of XRef subsections that can be then serialized by either
        :meth:`.write_standard_xref_section` or :meth:`.write_compressed_xref_section`.
        """
        rows = sorted(rows, key=lambda entry: entry[0])
        subsections: defaultdict[int, list[tuple[int, PdfXRefEntry]]] = defaultdict(list)

        first_obj_num = rows[0][0]

        for entry in rows:
            subsections[first_obj_num].append(entry)
            if len(subsections[first_obj_num]) <= 1:
                continue

            first_key, _ = subsections[first_obj_num][-1]
            second_key, _ = subsections[first_obj_num][-2]

            if first_key != second_key and abs(first_key - second_key) != 1:
                # The keys should belong in different subsections. Move the last key to a
                # different subsection and set that subsection for the rest of entries.
                last = subsections[first_obj_num].pop()
                first_obj_num = last[0]
                subsections[first_obj_num].append(last)

        return [
            PdfXRefSubsection(obj_num, len(entries), [ent for _, ent in entries])
            for obj_num, entries in subsections.items()
        ]

    def write_standard_xref_section(self, subsections: list[PdfXRefSubsection]) -> int:
        """Appends a standard XRef section (see § 7.5.4, "Cross-Reference Table") to the document.
        Returns the ``startxref`` offset that should be written to the document."""
        startxref = len(self.content)
        self.content_lines.append(b"xref" + self.eol)

        for subsection in subsections:
            self.content_lines.append(
                f"{subsection.first_obj_number} {subsection.count}".encode() + self.eol
            )

            for entry in subsection.entries:
                if isinstance(entry, InUseXRefEntry):
                    self.content_lines.append(
                        f"{entry.offset:0>10} {entry.generation:0>5} n".encode()
                    )
                elif isinstance(entry, FreeXRefEntry):
                    self.content_lines.append(
                        f"{entry.next_free_object:0>10} {entry.gen_if_used_again:0>5} f".encode()
                    )
                else:
                    raise PdfWriteError(
                        "Cannot write a compressed XRef entry within a standard XRef section."
                    )

                self.content_lines.append(self.eol)

        return startxref

    def write_compressed_xref_section(self, section: PdfXRefSection) -> int:
        """Appends a compressed XRef stream (see § 7.5.8, "Cross-Reference Streams") from
        ``section`` (to use as part of the extent) to the document.

        Returns the ``startxref`` offset that should be written to the document."""

        indices: PdfArray[PdfArray[int]] = PdfArray()
        table_rows: list[list[int]] = []

        for subsection in section.subsections:
            indices.append(PdfArray([subsection.first_obj_number, subsection.count]))

            for entry in subsection.entries:
                if isinstance(entry, FreeXRefEntry):
                    table_rows.append([0, entry.next_free_object, entry.gen_if_used_again])
                elif isinstance(entry, InUseXRefEntry):
                    table_rows.append([1, entry.offset, entry.generation])
                elif isinstance(entry, CompressedXRefEntry):
                    table_rows.append([2, entry.objstm_number, entry.index_within])

        def max_width(col: tuple[int, ...]) -> int:
            return ((max(col).bit_length() + 7) // 8) or 1

        widths = [max_width(column) for column in zip(*table_rows)]
        content_per_row = []
        for row in table_rows:
            content_per_row.append(
                b"".join(item.to_bytes(widths[idx], "big") for idx, item in enumerate(row))
            )

        contents = b"".join(content_per_row)

        stream = PdfStream(
            PdfDictionary(
                Type=PdfName(b"XRef"),
                W=PdfArray(widths),
                Index=PdfArray(sum(indices, start=PdfArray())),
                Length=len(contents),
                **section.trailer.data,
            ),
            contents,
        )

        highest_objnum = sum(max(indices, key=sum))
        return self.write_object((highest_objnum, 0), stream)

    def write_trailer(
        self, trailer: PdfDictionary | None = None, startxref: int | None = None
    ) -> None:
        """Appends a standard ``trailer`` to the document (see § 7.5.5, "File Trailer")
        alongside the ``startxref`` offset.

        Both arguments are optional, indicating their presence in the appended output.
        If the XRef section written previously was an XRef stream, the trailer has
        already been written and should be ``None``.
        """
        if trailer is not None:
            self.content_lines.append(b"trailer" + self.eol)
            self.content_lines.append(serialize_dictionary(trailer) + self.eol)

        if startxref is not None:
            self.content_lines.append(b"startxref" + self.eol)
            self.content_lines.append(str(startxref).encode() + self.eol)

    def write_eof(self) -> None:
        """Appends the End-Of-File marker to the document."""
        self.content_lines.append(b"%%EOF" + self.eol)



from __future__ import annotations

import enum
from typing import Annotated, Literal, Union, cast

from typing_extensions import Self

from ..common.dictmodels import defaultize, dictmodel, field
from ..cos.objects.base import PdfName
from ..cos.objects.containers import PdfArray, PdfDictionary

PageLayout = Literal[
    "SinglePage", "OneColumn", "TwoColumnLeft", "TwoColumnRight", "TwoPageLeft", "TwoPageRight"
]
PageMode = Literal["UseNone", "UseOutlines", "UseThumbs", "FullScreen", "UseOC", "UseAttachments"]
NonFSPageMode = Literal["UseNone", "UseOutlines", "UseThumbs", "UseOC"]
TextDirection = Literal["L2R", "R2L"]
PageBoundary = Literal["MediaBox", "CropBox", "BleedBox", "TrimBox", "ArtBox"]
PrintScalingMode = Literal["None", "AppDefault"]
PaperHandlingOption = Literal["Simplex", "DuplexFlipShortEdge", "DuplexFlipLongEdge"]
Enforceable = Literal["PrintScaling"]


class UserAccessPermissions(enum.IntFlag):
    """User access permissions as specified in the P entry of the document's standard
    encryption dictionary.

    See "Table 22 - Standard security handler user access permissions" in the PDF spec
    for details.
    """

    # bits 0 and 1 are reserved and should be 0.

    PRINT = 1 << 2
    """For security revision 2 or greater, Print the document. If the document uses 
    revision 3 or greater, print quality may be influenced by :attr:`.FAITHFUL_PRINT`."""

    MODIFY = 1 << 3
    """Modify the contents of the document. May be influenced by :attr:`.MANAGE_ANNOTATIONS`, 
    :attr:`.FILL_FORM_FIELDS`, and :attr:`.ASSEMBLE_DOCUMENT`."""

    COPY_CONTENT = 1 << 4
    """Copy or extract text and graphics. Assistive technology should assume this bit
    as set for its purposes, as per :attr:`.ACCESSIBILITY`."""

    MANAGE_ANNOTATIONS = 1 << 5
    """Add or modify text annotations, fill interactive form fields and, depending 
    on whether :attr:`.MODIFY` is set, create and modify form fields."""

    # bits 6 and 7 are reversed and should be 1.

    FILL_FORM_FIELDS = 1 << 8
    """For security revision 3 or greater, fill existing interactive form 
    fields, even if :attr:`.MANAGE_ANNOTATIONS` is clear."""

    ACCESSIBILITY = 1 << 9
    """(deprecated in PDF 2.0) Extract content for the purposes of accessibility.
    
    This bit should always be set for compatibility with processors supporting 
    earlier specifications.
    """

    ASSEMBLE_DOCUMENT = 1 << 10
    """For security revision 3 or greater, assemble the document (i.e. insert, rotate, 
    and delete pages, create outlines, etc.), even if :attr:`.MODIFY` is clear."""

    FAITHFUL_PRINT = 1 << 11  # revision 3+
    """For security revision 3 or greater, print the document in such a way that a 
    faithful digital representation of the PDF can be generated. 
    
    If this bit is not set (and :attr:`.PRINT` is set), printing shall be limited to 
    a low-level representation, possibly of lower quality.
    """

    # bits 12-31 are reserved and must be 1.


@dictmodel()
class ViewerPreferences(PdfDictionary):
    """The viewer preferences dictionary specifying the way a PDF viewer shall
    display a document on the screen.

    See § 12.2, "Viewer preferences" for details.
    """

    hide_toolbar: bool = field(default=False)
    """Whether to hide the interactive PDF processor's toolbars when the document is active."""

    hide_menubar: bool = field(default=False)
    """Whether to hide the interactive PDF processor's menubar when the document is active."""

    hide_window_ui: bool = field("HideWindowUI", default=False)
    """Whether to hide UI elements in the document's window (such as scroll bars or
    navigation controls), leaving only the document's contents displayed."""

    fit_window: bool = field(default=False)
    """Whether to resize the document's window to fit the size of the page."""

    center_window: bool = field(default=False)
    """Whether to center the document's window position on the screen."""

    display_doc_title: bool = field(default=False)
    """(PDF 1.4) Whether the document's window title should display the title described in 
    the document's metadata. If False, the title bar should instead display the name of the 
    PDF file containing the document."""

    non_full_screen_page_mode: NonFSPageMode = field(default="UseNone")
    """The document's page mode displayed when exiting full-screen mode. This property is only
    relevant if the PageMode entry in the catalog is set to 'FullScreen' and should be ignored
    otherwise. Accepted values are 'UseNone', 'UseOutlines', 'UseThumbs', and 'UseOC'."""

    direction: TextDirection = field(default="L2R")
    """The predominant logical content order for text. Either 'L2R' (left to right, default)
    or 'R2L' (right to left). This is effectively a display hint and has no direct effect
    on the contents of the document."""

    view_area: PageBoundary = field(default="CropBox")
    """(deprecated in PDF 2.0) The name of the page boundary representing the area of a page
    that shall be displayed when viewing the document on the screen. The value should be the
    key of the relevant page boundary in a page object. If no such boundary is defined,
    the default value ('CropBox') is used.

    Accepted values are 'CropBox', 'MediaBox', 'BleedBox', 'TrimBox', and 'ArtBox'.
    """

    view_clip: PageBoundary = field(default="CropBox")
    """(deprecated in PDF 2.0) The name of the page boundary representing to which the
    contents of a page shall be clipped when viewing the document. Similar to ViewArea,
    the value should be the key of the relevant page boundary in a page object.
    """

    print_area: PageBoundary = field(default="CropBox")
    """(deprecated in PDF 2.0) The name of the page boundary representing the area of a
    page that shall be rendered when printing the document. Similar to ViewArea, the value
    should be the key of the relevant page boundary in a page object."""

    print_clip: PageBoundary = field(default="CropBox")
    """(deprecated in PDF 2.0) The name of the page boundary representing to which the
    contents of a page shall be clipped when printing the document. Similar to ViewArea,
    the value should be the key of the relevant page boundary in a page object."""

    print_scaling: PrintScalingMode = field(default="AppDefault")
    """The page scaling option to select when a print dialog is displayed for this document.

    Accepted values are 'None' meaning no page scaling or 'AppDefault' (default) indicating
    that the interactive PDF processor should select its default print scaling value."""

    duplex: Union[PaperHandlingOption, None] = field(default=None)
    """The paper handling option to use when printing the document. Should be either of:

    - Simplex: Print single-sided
    - DuplexFlipShortEdge: Duplex, flip on the short edge of the sheet
    - DuplexFlipLongEdge: Duplex, flip on the long edge of the sheet

    If this value is none, the document producer may choose their own default setting.
    """

    pick_tray_by_pdf_size: Union[bool, None] = field(default=None)
    """Whether the PDF page size shall be used to select the input paper tray. This setting
    influences only the preset values used to populate the print dialog. This setting
    has no effect on systems that do not provide the ability to pick the input tray by size.

    If this value is none, the document producer may choose their own default setting.
    """

    print_page_range: Union[PdfArray[int], None] = field(default=None)
    """The page numbers used to initialize the print dialog box. The array should
    contain an even number of values interpreted as pairs, with each pair specifying
    the first and last pages in a sub-range of pages to be printed (the first page being
    denoted by the number 1).

    If this value is none, the document producer may choose their own default setting.
    """

    num_copies: Union[int, None] = field(default=None)
    """The number of copies that shall be printed when the print dialog is opened for this file.

    If this value is none, the document producer may choose their own default setting,
    though this setting is usually 1.
    """

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.data = mapping.data

        return dictionary

    @property
    def enforce(self) -> list[Enforceable] | None:
        """(PDF 2.0) An array of names of viewer preferences that shall be enforced by
        PDF processors and that shall not be overridden by subsequent selections in
        the application user interface."""
        if "Enforce" not in self:
            return

        enforced = cast(PdfArray[PdfName], self["Enforce"])

        return [cast(Enforceable, it.value.decode()) for it in enforced]


@dictmodel()
class DeveloperExtension(PdfDictionary):
    """An entry in an extension dictionary (see § 7.12.3, "Developer extensions dictionary")."""

    base_version: Annotated[str, "name"]
    """The name of the PDF version to which this extension applies.

    The name shall be consistent with the syntax used for the Version entry
    of the catalog dictionary (see § 7.7.2, "Document catalog dictionary").
    """

    level: int = field("ExtensionLevel")
    """An developer-defined integer denoting the extension being used.

    If the developer introduces more than one extension to a given base version,
    the extension level assigned by the developer should increase over time.
    """

    url: Union[str, None] = field("URL", default=None)
    """(PDF 2.0) A URL referring to the documentation for this extension."""

    revision: Union[str, None] = field("ExtensionRevision", default=None)
    """(PDF 2.0) Additional revision information on the extension level being used."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = defaultize(cls)
        dictionary.data = mapping.data

        return dictionary


class ExtensionMap(PdfDictionary):
    """A map defining developer extensions in a document (see § 7.12, "Extensions dictionary")."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.data = mapping.data

        return dictionary

    def query(self, key: str) -> DeveloperExtension | list[DeveloperExtension]:
        """Returns a developer-defined extension (or a sequence of them) for
        a base prefix ``key``."""

        if key == "Type":
            raise ValueError("not a valid extension name")

        extension = self[key]
        if isinstance(extension, PdfArray):
            return [DeveloperExtension.from_dict(cast(PdfDictionary, ext)) for ext in extension]

        return DeveloperExtension.from_dict(cast(PdfDictionary, extension))


@dictmodel()
class MarkInfo(PdfDictionary):
    """Information relevant to specialized uses of structured PDF documents.

    See § 14.7, "Logical structure" for details.
    """

    marked: bool = False
    """Whether the document claims to conform to tagged PDF conventions."""

    suspects: bool = False
    """(PDF 1.6; deprecated in PDF 2.0) Whether the document includes tag suspects
    which are applied for marked content elements whose page content order could not
    be determined.
    
    In such case, the document may not fully conform to tagged PDF conventions.
    """

    user_properties: bool = False
    """(PDF 1.6) Whether structure elements including user properties are present 
    in the document.
    
    See § 14.7.6.4, "User properties" for details.
    """

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = cls()
        dictionary.data = mapping.data

        return dictionary



from __future__ import annotations

import enum
from typing import Generator, Literal, Union, cast

from typing_extensions import Self

from ..common.dictmodels import defaultize, dictmodel, field
from ..cos.objects.base import PdfName, PdfReference
from ..cos.objects.containers import PdfArray, PdfDictionary
from ..cos.objects.stream import PdfStream
from ..cos.tokenizer import ContentStreamTokenizer

AnnotationKind = Literal[
    "Text",
    "Link",
    "FreeText",
    "Line",
    "Square",
    "Circle",
    "Polygon",
    "PolyLine",
    "Highlight",
    "Underline",
    "Squiggly",
    "StrikeOut",
    "Caret",
    "Stamp",
    "Ink",
    "Popup",
    "FileAttachment",
    "Sound",
    "Movie",
    "Screen",
    "Widget",
    "PrinterMark",
    "TrapNet",
    "Watermark",
    "3D",
    "Redact",
    "Projection",
    "RichMedia",
]

TabOrder = Literal["R", "C", "S", "A", "W"]


class AnnotationFlags(enum.IntFlag):
    """Flags for a particular annotation. See § 12.5.3, "Annotation flags" for details."""

    NULL = 0
    """A default value meaning that no flags are set."""

    INVISIBLE = 1 << 0
    """If the annotation is non-standard, do not render or print the annotation.
    
    If this flag is clear, the annotation shall be rendered according to its 
    appearance stream.
    """

    HIDDEN = 1 << 1
    """Do not render the annotation or allow user interaction with it."""

    PRINT = 1 << 2
    """Print the annotation when the page is printed unless :attr:`.AnnotationFlags.HIDDEN` 
    is set. If clear, do not print the annotation."""

    NO_ZOOM = 1 << 3
    """Do not scale the annotation's appearance to the page's zoom factor."""

    NO_ROTATE = 1 << 4
    """Do not rotate the annotation to match the page's rotation."""

    NO_VIEW = 1 << 5
    """Do not render the annotation or allow user interaction with it, but still
    allow printing according to the :attr:`.AnnotationFlags.PRINT` flag."""

    READ_ONLY = 1 << 6
    """Do not allow user interaction with the annotation. This is ignored for Widget
    annotations."""

    LOCKED = 1 << 7
    """Do not allow the annotation to be removed or its properties to be modified
    but still allow its contents to be modified."""

    TOGGLE_NO_VIEW = 1 << 8
    """Toggle the :attr:`.AnnotationFlags.No_VIEW` flag when selecting or hovering 
    over the annotation."""

    LOCKED_CONTENTS = 1 << 9
    """Do not allow the contents of the annotation to be modified."""


@dictmodel()
class Annotation(PdfDictionary):
    """An annotation associates an object such as a note, link, or multimedia element
    with a location on a page of a PDF document (see § 12.5, "Annotations")."""

    kind: AnnotationKind = field("Subtype")
    """The kind of annotation. See "Table 171: Annotation types" in the PDF spec 
    for an overview of their functions."""

    rect: PdfArray[float]
    """A rectangle specifying the location of the annotation in the page."""

    contents: str
    """The text contents that shall be displayed when the annotation is open or, if this
    annotation kind does not display text, an alternate description of the annotation's 
    contents."""

    name: str = field("NM")
    """An annotation name uniquely identifying the annotation among others in its page."""

    last_modified: Union[str, None] = field("M", default=None)
    """The date and time the annotation was most recently modified. This value should
    be a PDF date string but PDF processors are expected to accept and display a string
    in any format."""

    language: Union[str, None] = field("Lang", default=None)
    """(PDF 2.0) A language identifier specifying the natural language for all 
    text in the annotation except where overridden by other explicit language 
    specifications (see § 14.9.2, "Natural language specification")."""

    flags: AnnotationFlags = field("F", default=AnnotationFlags.NULL.value)
    """Flags specifying various characteristics of the annotation."""

    color: Union[float, None] = field("C", default=None)
    """An array of 0 to 4 numbers in the range 0.0 to 1.0, representing a color used
    for the following purposes:

    - The background of the annotation's icon when closed.
    - The title bar of the annotation's popup window.
    - The border of a link annotation.

    The number of array elements determines the color space in which the color shall
    be defined: 0 is no color or transparent; 1 is grayscale; 3 is RGB; and 4 is CMYK.
    """

    @classmethod
    def from_dict(cls, mapping: PdfDictionary) -> Self:
        dictionary = defaultize(cls)
        dictionary.data = mapping.data

        return dictionary


@dictmodel(init=False)
class Page(PdfDictionary):
    """A page in a PDF document (see § 7.7.3.3, "Page objects").

    Arguments:
        size (tuple[float, float]):
            The width and height of the physical medium in which the page should
            be printed or displayed. Values shall be provided in multiples of
            1/72 of an inch.

        indirect_ref (PdfReference, optional):
            The indirect reference that this page object is referred to by.

            In typical usage, this value need not be specified.
            pdfnaut will take care of populating it.
    """

    mediabox: PdfArray[float] = field("MediaBox")
    """A rectangle defining the boundaries of the physical medium in which the page
    should be printed or displayed."""

    cropbox: Union[PdfArray[float], None] = field("CropBox", default=None)
    """A rectangle defining the visible region of the page.
    
    If none, the cropbox is the same as the mediabox.
    """

    bleedbox: Union[PdfArray[float], None] = field("BleedBox", default=None)
    """A rectangle defining the region to which the contents of the page shall be 
    clipped when output in a production environment.
    
    If none, the bleedbox is the same as the cropbox.
    """

    trimbox: Union[PdfArray[float], None] = field("TrimBox", default=None)
    """A rectangle defining the intended dimensions of the finished page after trimming.

    If none, the trimbox is the same as the cropbox.
    """

    artbox: Union[PdfArray[float], None] = field("ArtBox", default=None)
    """A rectangle defining the extent of the page's meaningful content as intended 
    by the page's creator.
    
    If none, the artbox is the same as the cropbox.
    """

    resources: Union[PdfDictionary, None] = None
    """Resources required by the page contents.

    If the page requires no resources, this should return an empty resource
    dictionary. If the page inherits its resources from an ancestor,
    this should return None.
    """

    tab_order: Union[TabOrder, None] = field("Tabs", default=None)
    """(optional; PDF 1.5) The tab order to be used for annotations on the page.
    If present, it shall be one of the following values:

    - R: Row order
    - C: Column order
    - S: Logical structure order
    - A: Annotations array order (PDF 2.0)
    - W: Widget order (PDF 2.0)
    """

    user_unit: float = 1
    """The size of a user space unit, in multiples of 1/72 of an inch (by default, 1)."""

    rotation: int = field("Rotate", default=0)
    """The number of degrees by which the page shall be visually rotated clockwise.
    The value is a multiple of 90 (by default, 0)."""

    metadata: Union[PdfStream, None] = None
    """A metadata stream, generally written in XMP, containing information about this page."""

    @classmethod
    def from_dict(cls, mapping: PdfDictionary, indirect_ref: PdfReference | None = None) -> Self:
        dictionary = cls(size=(0, 0), indirect_ref=indirect_ref)
        dictionary.data = mapping.data

        return dictionary

    def __init__(
        self, size: tuple[float, float], *, indirect_ref: PdfReference | None = None
    ) -> None:
        super().__init__()

        self.indirect_ref = indirect_ref

        self["Type"] = PdfName(b"Page")
        self["MediaBox"] = PdfArray([0, 0, *size])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} mediabox={self.mediabox!r} rotation={self.rotation!r}>"

    @property
    def content_stream(self) -> ContentStreamTokenizer | None:
        """An iterator over the instructions producing the contents of this page."""
        if "Contents" not in self:
            return

        contents = cast("PdfStream | PdfArray[PdfStream]", self["Contents"])

        if isinstance(contents, PdfArray):
            # when Contents is an array, it shall be concatenated into a single
            # content stream with at least one whitespace character in between.
            return ContentStreamTokenizer(b"\n".join(stm.decode() for stm in contents))

        return ContentStreamTokenizer(contents.decode())

    @property
    def annotations(self) -> Generator[Annotation, None, None]:
        """All annotations associated with this page (see § 12.5, "Annotations"
        and :class:`.Annotation`)."""
        for annot in cast(PdfArray[PdfDictionary], self.get("Annots", PdfArray())):
            yield Annotation.from_dict(annot)



from __future__ import annotations

from .catalog import (
    PageLayout,
    PageMode,
    UserAccessPermissions,
    ViewerPreferences,
)
from .page import Annotation, AnnotationFlags, Page
from .trailer import Info
from .xmp import XmpMetadata

__all__ = (
    "PageLayout",
    "PageMode",
    "Page",
    "Annotation",
    "AnnotationFlags",
    "Info",
    "UserAccessPermissions",
    "ViewerPreferences",
    "XmpMetadata",
)



from __future__ import annotations

import datetime
import enum
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    Literal,
    Protocol,
    TypeVar,
    Union,
    cast,
    get_origin,
)

from typing_extensions import get_args

from ..common.dates import encode_iso8824, parse_iso8824
from ..cos.objects.base import (
    PdfHexString,
    PdfName,
    PdfObject,
    encode_text_string,
    parse_text_string,
)
from ..cos.objects.containers import PdfDictionary

if TYPE_CHECKING:
    from .dictmodels import Field


MISSING = type("MISSING", (), {})()


class Accessor(Protocol):
    field: Field

    def __init__(self, field: Field) -> None: ...
    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> Any: ...
    def __set__(self, obj: PdfDictionary, value: Any) -> None: ...
    def __delete__(self, obj: PdfDictionary) -> None: ...


class StandardAccessor:
    """An accessor defining a key whose value is a type that does not require
    a complex mapping such as booleans, numbers, and certain name objects.

    Text strings and dates have special handling and are better served by the
    :class:`.TextStringAccessor` and :class:`.DateAccessor` classes respectively.
    """

    def __init__(self, field: Field) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> PdfObject:
        if self.field.default is MISSING:
            return obj[self.field.key]

        return obj.get(self.field.key, self.field.default)

    def __set__(self, obj: PdfDictionary, value: PdfObject | None) -> None:
        if value is None:
            return self.__delete__(obj)

        obj[self.field.key] = value

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field.key, None)


class NameAccessor:
    """An accessor defining a key whose value may be any of a set of names."""

    def __init__(self, field: Field) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> str | None:
        if self.field.default is MISSING:
            return cast(PdfName, obj[self.field.key]).value.decode()

        if self.field.default is None:
            default = None
        else:
            default = PdfName(self.field.default.encode())

        name = obj.get(self.field.key, default)
        if isinstance(name, PdfName):
            return name.value.decode()

    def __set__(self, obj: PdfDictionary, value: str | None) -> None:
        if value is None:
            return self.__delete__(obj)

        obj[self.field.key] = PdfName(value.encode())

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field.key, None)


class TextStringAccessor:
    """An accessor defining a key whose value is a text string (``§ 7.9.2.2 Text string type``)."""

    def __init__(self, field: Field) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> str | None:
        if (value := obj.get(self.field.key)) is not None:
            return parse_text_string(cast("PdfHexString | bytes", value))
        return self.field.default

    def __set__(self, obj: PdfDictionary, value: str | None) -> None:
        if value is None:
            return self.__delete__(obj)

        obj[self.field.key] = encode_text_string(value)

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field.key, None)


class DateAccessor:
    """An accessor defining a key whose value is a date (see § 7.9.4. "Dates")."""

    def __init__(self, field: Field) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> datetime.datetime | None:
        text = TextStringAccessor(self.field).__get__(obj)

        if text is not None:
            return parse_iso8824(text)
        return self.field.default

    def __set__(self, obj: PdfDictionary, value: datetime.datetime | None) -> None:
        if value is None:
            return self.__delete__(obj)

        TextStringAccessor(self.field).__set__(obj, encode_iso8824(value))

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field.key, None)


E = TypeVar("E", bound=enum.IntFlag)


class BitFlagAccessor(Generic[E]):
    """An accessor defining a key whose value is part of a set of bit flags."""

    def __init__(self, field: Field) -> None:
        self.field = field

    def __get__(self, obj: PdfDictionary, objtype: Any | None = None) -> E | None:
        assert self.field.metadata and "enum_cls" in self.field.metadata

        value = obj.get(self.field.key, self.field.default)
        if value is not None:
            return self.field.metadata["enum_cls"](value)

    def __set__(self, obj: PdfDictionary, value: E | None) -> None:
        if value is None:
            return self.__delete__(obj)

        obj[self.field.key] = int(value)

    def __delete__(self, obj: PdfDictionary) -> None:
        obj.pop(self.field.key, None)


def lookup_accessor(value_type: type) -> tuple[type[Accessor], dict[str, Any]]:
    if value_type is str:
        return TextStringAccessor, {}
    elif value_type is datetime.datetime:
        return DateAccessor, {}
    elif get_origin(value_type) is Annotated:
        type_, subtype, *_ = get_args(value_type)
        if type_ is str:
            if subtype.lower() == "text":
                return TextStringAccessor, {}
            elif subtype.lower() == "name":
                return NameAccessor, {}
            else:
                raise TypeError(f"{subtype!r} not a valid subtype for a string accessor")

        raise NotImplementedError(f"accessor from annotated form {value_type!r} not implemented")
    elif get_origin(value_type) is Union:
        args = get_args(value_type)
        assert len(args) >= 1

        if len(args) > 2:
            raise ValueError(f"cannot create accessor for type {value_type!r}")

        if isinstance(args[-1], type(None)):
            raise NotImplementedError("only supported union form is Union[T, None]")

        return lookup_accessor(args[0])
    elif get_origin(value_type) is Literal:
        return NameAccessor, {}
    elif isinstance(value_type, enum.IntFlag):
        return BitFlagAccessor[value_type], {"enum_cls": value_type}

    return StandardAccessor, {}



from __future__ import annotations

from inspect import getattr_static
from typing import Any, TypeVar, cast

from typing_extensions import dataclass_transform, get_type_hints

from ..cos.objects.containers import PdfDictionary
from .accessors import MISSING, Accessor, lookup_accessor

_T = TypeVar("_T")


class Field:
    def __init__(
        self,
        key: str | None = None,
        default: Any = MISSING,
        init: bool | None = None,
        repr_: bool | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        # To be filled in later
        self.name: str | None = None
        self.type_: type | None = None

        self._key = key
        self.default = default
        self.init = init
        self.repr_ = repr_
        self.metadata = metadata

    @property
    def key(self) -> str:
        if self._key is None:
            raise ValueError(f"No key assigned for field {self.name!r}.")

        return self._key


def field(
    key: str | None = None,
    default: Any = MISSING,
    init: bool | None = None,
    repr_: bool | None = None,
    metadata: dict[str, Any] | None = None,
) -> Any:
    """Defines a field in a dictmodel.

    Arguments:
        key (str, optional):
            The name of the key that will be accessed by this field. If not specified,
            the key will be the title-cased version of the field name.

        default (Any, optional):
            The default value of the field if it is not specified. If no default is
            specified, the field is assumed to be required.

        init (bool | None, optional):
            Whether this field will appear as part of the class constructor. If not specified,
            it defaults to the value of the ``init`` argument in the dictmodel.

        repr_ (bool | None, optional):
            Whether this field will appear as part of the class representation.
            If not specified, it defaults to the value of the ``repr_`` argument
            in the dictmodel.

        metadata (dict[str, Any], optional):
            Additional metadata for this field which may be used by the accessor.
    """
    return Field(key, default, init, repr_, metadata)


T = TypeVar("T")


def defaultize(cls: type[T]) -> T:
    """Returns an instance of a dictmodel ``cls`` initialized with default values."""
    accessors = getattr(cls, "__accessors__", MISSING)
    if accessors is MISSING:
        raise TypeError(f"type {cls!r} is not a dictmodel")

    mapping: dict[str, Any] = {}
    for acc in cast(list[Accessor], accessors):
        assert acc.field.name is not None
        if not acc.field.init:
            continue

        if acc.field.default is not MISSING:
            mapping[acc.field.name] = acc.field.default
        else:
            mapping[acc.field.name] = None

    return cls(**mapping)


def snake_to_title_case(value: str) -> str:
    return "".join(val.title() for val in value.split("_"))


T = TypeVar("T")


def build_repr(cls: type[T], repr_accessors: list[Accessor]):
    def _repr(self: T) -> str:
        attrs = []

        for acc in repr_accessors:
            assert acc.field.name is not None

            value = getattr(self, acc.field.name, acc.field.default)
            if value == acc.field.default:
                continue

            attrs.append(f"{acc.field.name}={value!r}")

        return f"{cls.__name__}({', '.join(attrs)})"

    return _repr


def create_accessors(cls, *, parent_init: bool = True, parent_repr: bool = True) -> list[Accessor]:
    accessors = []

    for attr, type_ in get_type_hints(cls, include_extras=True).items():
        default = getattr_static(cls, attr, MISSING)

        if isinstance(default, Field):
            model_field = default
        elif hasattr(default, "field"):
            # inherited field from an accessor
            model_field = default.field
        else:
            model_field = Field(default=default)

        if model_field._key is None:
            model_field._key = snake_to_title_case(attr)

        model_field.name = attr
        model_field.type_ = type_
        if model_field.init is None:
            model_field.init = parent_init

        if model_field.repr_ is None:
            model_field.repr_ = parent_repr

        accessor, metadata = lookup_accessor(type_)
        if accessor is None:
            raise ValueError(f"No accessor registered for type {type_!r}")

        if metadata is not None:
            if model_field.metadata is not None:
                model_field.metadata |= metadata
            else:
                model_field.metadata = metadata

        accessors.append(accessor(model_field))

    return accessors


@dataclass_transform(field_specifiers=(field,))
def dictmodel(*, init: bool = True, repr_: bool = True):
    def wrapper(cls: type[_T]) -> type[_T]:
        if not issubclass(cls, PdfDictionary):
            raise TypeError("cls must be a subclass of PdfDictionary")

        accessors = create_accessors(cls, parent_init=init, parent_repr=repr_)

        init_args = ["self"]

        for accessor in accessors:
            assert accessor.field.name is not None

            setattr(cls, accessor.field.name, accessor)

            if not accessor.field.init:
                continue

            init_arg_string = accessor.field.name
            if accessor.field.default is not MISSING:
                init_arg_string += f" = {accessor.field.default!r}"

            init_args.append(init_arg_string)

        required_subcls_args = []
        for acc in getattr(cls, "__accessors__", []):
            if not acc.field.init:
                continue

            if acc.field.default is not MISSING:
                required_subcls_args.append(f"{acc.field.name}={acc.field.name}")
            else:
                required_subcls_args.append(acc.field.name)

        init_fn_body = [
            f"def __init__({', '.join(init_args)}):",
            f"  super({cls.__name__}, self).__init__({', '.join(required_subcls_args)})",
        ]

        for acc in accessors:
            if not acc.field.init:
                continue

            init_fn_body.append(f"  self.{acc.field.name} = {acc.field.name}\n")

        repr_fn = build_repr(cls, [acc for acc in accessors if acc.field.repr_])
        namespace = {}

        exec("\n".join(init_fn_body), {cls.__name__: cls}, namespace)

        if "__init__" not in cls.__dict__:
            cls.__init__ = namespace["__init__"]
            cls.__init__.__name__ = "__init__"
            cls.__init__.__qualname__ = f"{cls.__qualname__}.__init__"

        if "__repr__" not in cls.__dict__:
            cls.__repr__ = repr_fn
            cls.__repr__.__name__ = "__repr__"
            cls.__repr__.__qualname__ = f"{cls.__qualname__}.__repr__"

        setattr(cls, "__accessors__", accessors)

        return cls

    return wrapper
